"""Module containing the DecimalComplex class which stores complex numbers with
``decimal.Decimal`` precision."""
from __future__ import annotations

import functools
from decimal import Decimal
from numbers import Complex
from typing import Callable, Union


def convert_input(func: Callable) -> Callable:
    """Converts inputs into valid ``DecimalComplex`` objects."""

    @functools.wraps(func)
    def convert_wrapper(self, c: Union[Decimal, Complex]) -> DecimalComplex:
        if isinstance(c, DecimalComplex):
            return func(self, c)
        if isinstance(c, Decimal):
            c = DecimalComplex(c)
            return func(self, c)
        if isinstance(c, Complex):
            c = DecimalComplex(str(c.real), str(c.imag))
            return func(self, c)

        raise TypeError("Must be of type numbers.Complex.")

    return convert_wrapper


class DecimalComplex(Complex):
    """Complex numbers represented by precision ``decimal.Decimal`` terms.

    Args:
        real (str, Decimal): the real part of the complex number. Defaults to
            ``0.0`` if not input.
        imag (str, Decimal): The imaginary part of the complex number. Defaults
            to ``0.0`` if not input.
    """

    def __init__(
        self, real: Union[str, Decimal] = "0.0", imag: Union[str, Decimal] = "0.0"
    ) -> None:
        self._real = Decimal(real)
        self._imag = Decimal(imag)

    @convert_input
    def __add__(self, c: Union[Decimal, Complex]) -> DecimalComplex:
        real = self.real + c.real
        imag = self.imag + c.imag
        return DecimalComplex(real, imag)

    def __radd__(self, c: Union[Decimal, Complex]) -> DecimalComplex:
        return self + c

    @convert_input
    def __sub__(self, c: Union[Decimal, Complex]) -> DecimalComplex:
        real = self.real - c.real
        imag = self.imag - c.imag
        return DecimalComplex(real, imag)

    def __rsub__(self, c: Union[Decimal, Complex]) -> DecimalComplex:
        return -(self - c)

    @convert_input
    def __mul__(self, c: Union[Decimal, Complex]) -> DecimalComplex:
        real = self.real * c.real - self.imag * c.imag
        imag = self.imag * c.real + self.real * c.imag
        return DecimalComplex(real, imag)

    def __rmul__(self, c: Union[Decimal, Complex]) -> DecimalComplex:
        return self * c

    # pylint: disable=missing-function-docstring
    @convert_input
    def __div__(self, c: Union[Decimal, Complex]) -> DecimalComplex:
        real = (self.real * c.real + self.imag * c.imag) / (c.real ** 2 + c.imag ** 2)
        imag = (self.imag * c.real - self.real * c.imag) / (c.real ** 2 + c.imag ** 2)
        return DecimalComplex(real, imag)

    # pylint: disable=missing-function-docstring
    @convert_input
    def __rdiv__(self, c: Union[Decimal, Complex]) -> DecimalComplex:
        return c / self

    @convert_input
    def __truediv__(self, c: Union[Decimal, Complex]) -> DecimalComplex:
        return self.__div__(c)

    @convert_input
    def __rtruediv__(self, c: Union[Decimal, Complex]) -> DecimalComplex:
        return c / self

    # pylint: disable=missing-function-docstring
    def __floordiv__(self, _: Union[Decimal, Complex]) -> DecimalComplex:
        raise TypeError("Cannot take floor of DecimalComplex")

    def __abs__(self) -> Decimal:
        return Decimal(self.real ** 2 + self.imag ** 2).sqrt()

    def __pos__(self) -> DecimalComplex:
        return self

    def __neg__(self) -> DecimalComplex:
        return DecimalComplex(-self.real, -self.imag)

    @convert_input
    def __eq__(self, c: DecimalComplex) -> bool:
        return self.real == c.real and self.imag == c.imag

    @convert_input
    def __ne__(self, c: DecimalComplex) -> bool:
        return not self.__eq__(c)

    def __str__(self) -> str:
        return f"{self.real}{self.imag:+}j"

    def __repr__(self):
        return f"DecimalComplex('{self.__str__()}')"

    def __pow__(self, c: Union[Decimal, Complex]) -> DecimalComplex:
        # TODO: calculate powers with precision using Decimal
        res = complex(self) ** complex(c)
        return DecimalComplex(str(res.real), str(res.imag))

    def __rpow__(self, c: Union[Decimal, Complex]) -> DecimalComplex:
        # TODO: calculate powers with precision using Decimal
        res = complex(c) ** complex(self)
        return DecimalComplex(str(res.real), str(res.imag))

    def __gt__(self, _: Union[Decimal, Complex]):
        self._not_implemented(">")

    def __lt__(self, _: Union[Decimal, Complex]):
        self._not_implemented("<")

    def __ge__(self, _: Union[Decimal, Complex]):
        self._not_implemented(">=")

    def __le__(self, _: Union[Decimal, Complex]):
        self._not_implemented("<=")

    def __bool__(self) -> bool:
        return self.real != 0 or self.imag != 0

    def __complex__(self) -> complex:
        return float(self.real) + float(self.imag) * 1j

    def _not_implemented(self, obs):
        """Raises a TypeError for the given (unimplemented) operation."""
        raise TypeError(f"Cannot use {obs} with complex numbers")

    def __hash__(self):
        return hash((self.real, self.imag))

    @property
    def real(self):
        """Return the real part of the complex number."""
        return self._real

    @property
    def imag(self):
        """Return the imaginary part of the complex number."""
        return self._imag

    def conjugate(self):
        """Return the complex conjugate of its argument."""
        return DecimalComplex(self.real, -self.imag)



"""This module contains the :class:`xir.Program` class and related classes."""
from __future__ import annotations

import re
import warnings
from decimal import Decimal
from typing import (
    Any,
    Collection,
    Dict,
    List,
    Mapping,
    MutableMapping,
    MutableSet,
    Optional,
    Sequence,
    Union,
)

from .decimal_complex import DecimalComplex
from .utils import strip

Wire = Union[int, str]
Param = Union[complex, str, Decimal, DecimalComplex, bool, List["Param"]]
Params = Union[List[Param], Dict[str, Param]]


def get_floats(params: Params) -> Params:
    """Converts `decimal.Decimal` and `DecimalComplex` objects to ``float`` and
    ``complex`` respectively"""
    if isinstance(params, List):
        params_with_floats = []
        for p in params:
            if isinstance(p, DecimalComplex):
                params_with_floats.append(complex(p))
            elif isinstance(p, Decimal):
                params_with_floats.append(float(p))
            elif isinstance(p, List):
                params_with_floats.append(get_floats(p))
            else:
                params_with_floats.append(p)

    elif isinstance(params, Dict):
        params_with_floats = {}
        for k, v in params.items():
            if isinstance(v, DecimalComplex):
                params_with_floats[k] = complex(v)
            elif isinstance(v, Decimal):
                params_with_floats[k] = float(v)
            elif isinstance(v, List):
                params_with_floats[k] = get_floats(v)
            else:
                params_with_floats[k] = v

    return params_with_floats


class Statement:
    """A general statement consisting of a name, optional parameters and wires.

    This is used for gate statements (e.g. ``rx(0.13) | [0]``) or output statements
    (e.g. ``sample(shots: 1000) | [0, 1]``).

    Args:
        name (str): name of the statement
        params (list, Dict): parameters for the statement (can be empty)
        wires (tuple): the wires on which the statement is applied

    Keyword args:
        inverse (bool): whether the statement is an inverse gate
        ctrl_wires (tuple): the control wires of a controlled gate statement
        use_floats (bool): Whether floats and complex types are returned instead of ``Decimal``
            and ``DecimalComplex`` objects. Defaults to ``True``.
    """

    def __init__(self, name: str, params: Params, wires: Sequence[Wire], **kwargs):
        self._name = name
        self._params = params
        self._wires = wires

        self._is_inverse = kwargs.get("inverse", False)
        self._ctrl_wires = kwargs.get("ctrl_wires", tuple())

        self._use_floats = kwargs.get("use_floats", True)

    def __str__(self):
        """Serialized string representation of a Statement."""
        if isinstance(self.params, dict):
            params = [f"{k}: {v}" for k, v in self.params.items()]
        else:
            params = [str(p) for p in self.params]
        params_str = ", ".join(params)
        if params_str != "":
            params_str = "(" + params_str + ")"

        wires = ", ".join([str(w) for w in self.wires])

        modifier_str = ""
        if len(self.ctrl_wires) != 0:
            ctrl_wires = ", ".join([str(w) for w in self.ctrl_wires])
            modifier_str = f"ctrl[{ctrl_wires}] "
        if self.is_inverse:
            modifier_str += "inv "

        return f"{modifier_str}{self.name}{params_str} | [{wires}]"

    @property
    def name(self) -> str:
        """Returns the name of the gate statement."""
        return self._name

    @property
    def params(self) -> Params:
        """Returns the parameters of the gate statement."""
        if self.use_floats:
            return get_floats(self._params)
        return self._params

    @property
    def wires(self) -> Sequence[Wire]:
        """Returns the wires that the gate is applied to."""
        return self._wires

    @property
    def is_inverse(self) -> bool:
        """Returns whether the statement applies an inverse gate."""
        return self._is_inverse

    @property
    def ctrl_wires(self) -> Sequence[Wire]:
        """Returns the control wires of a controlled gate statement.
        If no control wires are specified, an empty tuple is returned.
        """
        return self._ctrl_wires

    @property
    def use_floats(self) -> bool:
        """Returns whether floats and complex types are returned instead of
        ``Decimal`` and ``DecimalComplex`` objects, respectively.
        """
        return self._use_floats


class ObservableStmt:
    """Observable statements to be used in observable definitions.

    Args:
        pref (Decimal, int, str): prefactor to the observable terms
        terms (list): list of observables and the wire(s) they are applied to
    """

    def __init__(self, pref: Union[Decimal, int, str], terms: List, use_floats: bool = True):
        self._pref = pref
        self._terms = terms

        self._use_floats = use_floats

    def __str__(self) -> str:
        """Serialized string representation of an ObservableStmt."""
        terms = [f"{t[0]}[{t[1]}]" for t in self.terms]
        terms_as_string = " @ ".join(terms)
        pref = str(self.pref)

        return f"{pref}, {terms_as_string}"

    @property
    def pref(self) -> Union[Decimal, float, int, str]:
        """Returns the prefactor of this observable statement."""
        if isinstance(self._pref, Decimal) and self.use_floats:
            return float(self._pref)
        return self._pref

    @property
    def terms(self) -> List:
        """Returns the terms in this observable statement."""
        return self._terms

    @property
    def use_floats(self) -> bool:
        """Returns the float setting of this observable statement."""
        return self._use_floats

    @property
    def wires(self) -> Sequence[Wire]:
        """Returns the wires this observable statement is applied to."""
        return tuple(wires for _, wires in self.terms)


class Declaration:
    """General declaration for declaring observables, gates, functions and outputs.

    Args:
        name (str): name of the declaration
        type_ (str): The type of declaration. Can be either "gate", "obs", "out"
            or "func".
        params (Sequence[str]): parameters used by the declared object
        wires (Sequence[Wire]): wires that the declared object is applied to
    """

    def __init__(
        self,
        name: str,
        type_: str,
        params: Optional[Sequence[str]] = None,
        wires: Optional[Sequence[Wire]] = None,
    ) -> None:
        if type_ not in ("gate", "out", "obs", "func"):
            raise TypeError(f"Declaration type '{type_}' is invalid.")

        self._name = name
        self._type = type_
        self._params = list(params or [])
        self._wires = tuple(wires or ())

    def __str__(self) -> str:
        """Serialized string representation of a declaration."""
        wires = params = ""
        if self.params:
            params = "(" + ", ".join(map(str, self.params)) + ")"
        if self.wires:
            wires = "[" + ", ".join(map(str, self.wires)) + "]"

        return f"{self._type} {self.name}{params}{wires}"

    def __repr__(self) -> str:
        return f"<{self.type_.capitalize()} declaration:name={self.name}>"

    def __eq__(self, other):
        if isinstance(other, Declaration):
            self_data = self.name, self.type_, self.params, self.wires
            other_data = other.name, other.type_, other.params, other.wires
            return self_data == other_data
        return NotImplemented

    @property
    def name(self) -> str:
        """Returns the name of the declaration."""
        return self._name

    @property
    def type_(self) -> str:
        """Returns the type of the declaration."""
        return self._type

    @property
    def params(self) -> Sequence[str]:
        """Returns the parameters of the declaration."""
        return self._params

    @property
    def wires(self) -> Sequence[Wire]:
        """Returns the wires of the declaration."""
        return self._wires


class Program:
    """Structured representation of an XIR program.

    Args:
        version (str): Version number of the program. Must follow SemVer style (MAJOR.MINOR.PATCH).
        use_floats (bool): Whether floats and complex types are returned instead of ``Decimal``
            and ``DecimalComplex`` objects. Defaults to ``True``.
    """

    def __init__(self, version: str = "0.1.0", use_floats: bool = True):
        Program._validate_version(version)

        self._version = version
        self._use_floats = use_floats

        self._includes = []
        self._options = {}
        self._constants = {}
        self._statements = []

        self._declarations = {key: [] for key in ("gate", "func", "out", "obs")}

        self._gates = {}
        self._observables = {}

        self._variables = set()
        self._called_functions = set()

    def __repr__(self) -> str:
        """Returns a string representation of the XIR program."""
        return f"<Program: version={self._version}>"

    @property
    def version(self) -> str:
        """Returns the version number of the XIR program.

        Returns:
            str: version number
        """
        return self._version

    @property
    def use_floats(self) -> bool:
        """Returns whether floats and complex types are used by the XIR program.

        Returns:
            bool: whether floats and complex types are used
        """
        return self._use_floats

    @property
    def wires(self) -> Collection[Wire]:
        """Returns the wires of the XIR program.

        Returns:
            Collection[Wire]: collection of wires
        """
        wires = []
        for stmt in self.statements:
            wires.extend(stmt.wires)

        return set(wires)

    @property
    def called_functions(self) -> Collection[str]:
        """Returns the functions that are called in the XIR program.

        Returns:
            Collection[str]: collection of function names
        """
        return self._called_functions

    @property
    def declarations(self) -> Mapping[str, Sequence[Declaration]]:
        """Returns the declarations in the XIR program.

        Returns:
            Mapping[str, Sequence[Declaration]]: dictionary of declarations
            with the following keys: 'gate', 'func', 'out', and 'obs'
        """
        return self._declarations

    @property
    def gates(self) -> Mapping[str, Sequence[Statement]]:
        """Returns the gates in the XIR program.

        Returns:
            Mapping[str, Mapping[str, Sequence]]: dictionary of gates, each one
            consisting of a name and a dictionary with the following keys:
            'parameters', 'wires', and 'statements'
        """
        return self._gates

    @property
    def includes(self) -> Sequence[str]:
        """Returns the included XIR modules used by the XIR program.

        Returns:
            Sequence[str]: sequence of included XIR modules
        """
        return self._includes

    @property
    def observables(self) -> Mapping[str, Sequence[ObservableStmt]]:
        """Returns the observables in the XIR program.

        Returns:
            Mapping[str, Mapping[str, Sequence]]: dictionary of observables, each
            one consisting of a name and a dictionary with the following keys:
            'parameters', 'wires', and 'statements'
        """
        return self._observables

    @property
    def options(self) -> Mapping[str, Any]:
        """Returns the script-level options declared in the XIR program.

        Returns:
            Mapping[str, Any]: declared script-level options
        """
        if self.use_floats:
            options_with_floats = get_floats(self._options)
            # The following condition should always be True. For more context,
            # see https://github.com/XanaduAI/jet/pull/52/files#r681872696.
            if isinstance(options_with_floats, Dict):
                return options_with_floats
        return self._options

    @property
    def constants(self) -> Mapping[str, Any]:
        """Returns the script-level constants declared in the XIR program.

        Returns:
            Mapping[str, Any]: declared script-level constants
        """
        if self.use_floats:
            constants_with_floats = get_floats(self._constants)
            # The following condition should always be True. For more context,
            # see https://github.com/XanaduAI/jet/pull/52/files#r681872696.
            if isinstance(constants_with_floats, Dict):
                return constants_with_floats
        return self._constants

    @property
    def statements(self) -> Sequence[Statement]:
        """Returns the statements in the XIR program.

        Returns:
            Sequence[Statement]: sequence of statements
        """
        return self._statements

    @property
    def variables(self) -> Collection[str]:
        """Returns the free parameter variables used when defining gates and
        observables in the XIR program.

        Returns:
            Collection[str]: collection of free parameter variable names
        """
        return self._variables

    def search(self, decl_type: str, attr_type: str, name: str) -> Sequence:
        """Searches an XIR program for the wires or parameters of a declaration.

        Args:
            decl_type (str): type of the declaration ("gate", "obs", "out", or "func")
            attr_type (str): type of the attribute ("wires" or "params")
            name (str): name of the decalaration

        Returns:
            Sequence: wires or parameters of the declaration with the given name and type

        Raises:
            ValueError: if an invalid declaration or attribute type was provided
                or no declaration with the given name and type was found
        """
        decl_types = set(self.declarations.keys())
        if decl_type not in decl_types:
            raise ValueError(f"Declaration type '{decl_type}' must be one of {decl_types}.")

        attr_types = {"wires", "params"}
        if attr_type not in attr_types:
            raise ValueError(f"Attribute type '{attr_type}' must be one of {attr_types}.")

        decls = [decl for decl in self.declarations[decl_type] if decl.name == name]
        if not decls:
            raise ValueError(f"No {decl_type} declarations with the name '{name}' were found.")

        return getattr(decls[0], attr_type)

    def add_called_function(self, name: str) -> None:
        """Adds the name of a called function to the XIR program.

        Args:
            name (str): name of the function
        """
        self._called_functions.add(name)

    def add_declaration(self, decl: Declaration) -> None:
        """Adds a declaration to the XIR program.

        Args:
            decl (Declaration): the declaration
        """
        declarations = self._declarations[decl.type_]

        conflicts = [decl_ for decl_ in declarations if decl_.name == decl.name]
        if conflicts:
            warnings.warn(
                f"{decl.type_.title()} '{decl.name}' already declared."
                f"Replacing old declaration with new declaration."
            )

            conflict = conflicts[0]
            declarations.remove(conflict)

        declarations.append(decl)

    def add_gate(
        self,
        name: str,
        params: Sequence[str],
        wires: Sequence[Wire],
        statements: Sequence[Statement],
    ) -> None:
        """Adds a gate to the XIR program.

        Args:
            name (str): name of the gate
            params (Sequence[str]): parameters used in the gate
            wires (Sequence[Wire]): wires that the gate is applied to
            statements (Sequence[Statement]): statements that the gate applies
        """
        if name in self._gates:
            warnings.warn(
                f"Gate '{name}' already defined. Replacing old definition with new definition."
            )

        self.add_declaration(Declaration(name, "gate", params, wires))
        self._gates[name] = statements

    def add_include(self, include: str) -> None:
        """Adds an included XIR module to the XIR program.

        Args:
            include (str): name of the XIR module
        """
        if include in self._includes:
            warnings.warn(f"Module '{include}' is already included. Skipping include.")
            return

        self._includes.append(include)

    def add_observable(
        self,
        name: str,
        params: Sequence[str],
        wires: Sequence[Wire],
        statements: Sequence[ObservableStmt],
    ) -> None:
        """Adds an observable to the XIR program.

        Args:
            name (str): name of the observable
            params (Sequence[str]): parameters used in the observable
            wires (Sequence[Wire]): wires that the observable is applied to
            statements (Sequence[ObservableStmt]): statements that the observable applies
        """
        if name in self._observables:
            warnings.warn(
                f"Observable '{name}' already defined."
                "Replacing old definition with new definition."
            )

        self.add_declaration(Declaration(name, "obs", params, wires))
        self._observables[name] = statements

    def add_option(self, name: str, value: Any) -> None:
        """Adds an option to the XIR program.

        Args:
            name (str): name of the option
            value (Any): value of the option
        """
        if name in self._options:
            warnings.warn(f"Option '{name}' already set. Replacing old value with new value.")

        self._options[name] = value

    def add_constant(self, name: str, value: Any) -> None:
        """Adds a constant to the XIR program.

        Args:
            name (str): name of the constant
            value (Any): value of the constant
        """
        if name in self._constants:
            warnings.warn(f"Constant '{name}' already set. Replacing old value with new value.")

        self._constants[name] = value

    def add_statement(self, statement: Statement) -> None:
        """Adds a statement to the XIR program.

        Args:
            statement (Statement): the statement
        """
        self._statements.append(statement)

    def add_variable(self, name: str) -> None:
        """Adds the name of a free parameter variable to the XIR program.

        Args:
            name (str): name of the variable
        """
        self._variables.add(name)

    def clear_includes(self) -> None:
        """Clears the includes of an XIR program."""
        self._includes = []

    def serialize(self, minimize: bool = False) -> str:
        """Serializes an XIR program to an XIR script.

        Args:
            minimize (bool): whether to strip whitespace and newlines from file

        Returns:
            str: the serialized XIR script
        """
        res = []
        res.extend([f"use {use};" for use in self._includes])
        if len(self._includes) != 0:
            res.append("")

        if len(self.options) > 0:
            res.append("options:")
            res.extend([f"    {k}: {v};" for k, v in self.options.items()])
            res.append("end;\n")

        if len(self.constants) > 0:
            res.append("constants:")
            res.extend([f"    {k}: {v};" for k, v in self.constants.items()])
            res.append("end;\n")

        has_decl = False
        for v in self._declarations.values():
            for decl in v:
                declared_gate = decl.type_ == "gate" and decl.name in self.gates
                declared_obs = decl.type_ == "obs" and decl.name in self.observables
                if declared_gate or declared_obs:
                    continue

                res.append(f"{decl};")
                has_decl = True

        if has_decl:
            res.append("")

        for name, gate in self._gates.items():
            decl = [d for d in self.declarations["gate"] if d.name == name][0]

            res.append(str(decl) + ":")

            res.extend([f"    {stmt};" for stmt in gate])
            res.append("end;\n")

        for name, obs in self._observables.items():
            decl = [d for d in self.declarations["obs"] if d.name == name][0]

            res.append(str(decl) + ":")

            res.extend([f"    {stmt};" for stmt in obs])
            res.append("end;\n")

        res.extend([f"{str(stmt)};" for stmt in self._statements])

        res_script = "\n".join(res).strip()
        if minimize:
            return strip(res_script)
        return res_script

    @staticmethod
    def merge(*programs: Program) -> Program:
        """Merges one or more XIR programs into a new XIR program.

        The merged XIR program is formed by concatenating the given XIR programs
        in the order they are passed to the function. Warnings may be issued for
        duplicate declarations, gates, includes, observables, and options.

        Args:
            programs (Program): XIR programs to merge

        Returns:
            Program: the merged XIR program

        Raises:
            ValueError: if no XIR programs are provided or if at least two XIR
                programs have different versions or float settings
        """
        if len(programs) == 0:
            raise ValueError("Merging requires at least one XIR program.")

        version = programs[0].version
        if any(program.version != version for program in programs):
            raise ValueError("XIR programs with different versions cannot be merged.")

        use_floats = any(program.use_floats for program in programs)
        if any(program.use_floats != use_floats for program in programs):
            warnings.warn("XIR programs with different float settings are being merged.")

        result = Program(version=version, use_floats=use_floats)

        for program in programs:
            for called_function in program.called_functions:
                result.add_called_function(called_function)

            for name, stmts in program.gates.items():
                decl = next(d for d in program.declarations["gate"] if d.name == name)
                result.add_gate(name, decl.params, decl.wires, stmts)

            for include in program.includes:
                result.add_include(include)

            for name, stmts in program.observables.items():
                decl = [d for d in program.declarations["obs"] if d.name == name][0]
                result.add_observable(name, decl.params, decl.wires, stmts)

            for type_, decls in program.declarations.items():
                for decl in decls:
                    if type_ == "gate" and decl.name not in result.gates:
                        result.add_declaration(decl)
                    if type_ == "obs" and decl.name not in result.observables:
                        result.add_declaration(decl)
                    if type_ in ("func", "out"):
                        result.add_declaration(decl)

            for name, value in program.options.items():
                result.add_option(name, value)

            for statement in program.statements:
                result.add_statement(statement)

            for variable in program.variables:
                result.add_variable(variable)

        return result

    @staticmethod
    def resolve(library: Mapping[str, Program], name: str) -> Program:
        """Resolves the includes of an XIR program using an XIR library.

        Args:
            library (Mapping[str, Program]): mapping from names to XIR programs
            name (str): name of the root XIR program to resolve

        Returns:
            Program: XIR program obtained by recursively merging each XIR
            program starting from the root XIR program

        **Example**

        Consider an XIR program ``main`` which includes another XIR program ``util``:

        .. code-block:: python

            import xir

            # Create two simple XIR programs.
            main_program = xir.parse_script("use util; H2 | [0, 1];")
            util_program = xir.parse_script("gate H2, 0, 2; gate H2: H | [0]; H | [1]; end;")

        The ``main`` XIR program can be resolved using

        .. code-block:: python

            # Define a library containing the XIR programs in the resolution scope.
            library = {"main": main_program, "util": util_program}

            # Resolve the imports from the main XIR program.
            resolved_main_program = xir.Program.resolve(library, "main")

            # Display the result.
            print(resolved_main_program.serialize())

        The output is

        .. code-block:: none

            gate H2:
                H | [0];
                H | [1];
            end;

            H2 | [0, 1];
        """
        resolved_names = Program._resolve_include_order(library, name, stack=set(), cache={})
        resolved_programs = map(library.__getitem__, resolved_names)

        resolved_program = Program.merge(*resolved_programs)
        resolved_program.clear_includes()
        return resolved_program

    @staticmethod
    def _resolve_include_order(
        library: Mapping[str, Program],
        name: str,
        stack: MutableSet[str],
        cache: MutableMapping[str, Sequence[str]],
    ) -> Sequence[str]:
        """
        Resolves the include order of an XIR program using C3 superclass linearization.

        See https://en.wikipedia.org/wiki/C3_linearization for a summary of the
        properties that hold following the linearization.

        Args:
            library (Mapping[str, Program]): mapping from names to XIR programs
            name (str): name of the root XIR program to resolve
            stack (MutableSet[str]): names in the current call stack; prevents
                infinite recursion in the case of circular dependencies
            cache (MutableMapping[str, Sequence[str]]): mapping from names to
                linearizations; improves performance when an XIR program is
                included by multiple XIR programs

        Returns:
            Sequence[str]: names representing a C3 linearization of the given XIR program.

        Raises:
            KeyError: if ``library`` does not have an entry for ``name``, or the
                XIR program associated with ``name`` (transitively) includes
                another XIR program which does not have an entry in ``library``
            ValueError: if ``name`` is already in ``stack``, or the XIR program
                associated with ``name`` (transitively) includes an XIR program
                with a circular dependency
        """
        if name not in library:
            raise KeyError(f"XIR program '{name}' cannot be found.")

        if name in stack:
            raise ValueError(f"XIR program '{name}' has a circular dependency.")

        # The stack is only empty when processing the root XIR program.
        if stack and library[name].statements:
            raise ValueError(f"XIR program '{name}' contains a statement.")

        if name in cache:
            return cache[name]

        # Step 1: Generate the C3 linearizations of the included XIR programs.
        stack.add(name)

        linearizations = []
        for include in library[name].includes:
            # Call the list constructor to avoid modifying cached linearizations.
            linearization = list(Program._resolve_include_order(library, include, stack, cache))
            linearizations.append(linearization)

        stack.discard(name)

        # Step 2: Iteratively select the next XIR program to be included.
        includes = []

        while linearizations:
            tails = {include for include in includes[1:] for includes in linearizations}
            heads = (includes[0] for includes in linearizations)

            selection = next(head for head in heads if head not in tails)
            includes.append(selection)

            # Remove the included XIR program from the linearizations and delete
            # any empty linearizations that are generated as a result.
            for linearization in linearizations:
                if selection in linearization:
                    linearization.remove(selection)
            linearizations = [linearization for linearization in linearizations if linearization]

        # Unlike MRO, the current XIR program has lower precendence than its includes.
        includes.append(name)

        cache[name] = includes

        return includes

    @staticmethod
    def _validate_version(version: str) -> None:
        """Validates the given version number.

        Raises:
            TypeError: if the version number is not a string
            ValueError: if the version number is not a semantic version
        """
        if not isinstance(version, str):
            raise TypeError(f"Version '{version}' must be a string.")

        valid_match = re.fullmatch(r"\d+\.\d+\.\d+", version)
        if valid_match is None:
            raise ValueError(f"Version '{version}' must be a semantic version (MAJOR.MINOR.PATCH).")



"""This module contains functions for validating an XIR program"""

import re
from decimal import Decimal
from itertools import chain
from typing import MutableSet, Optional, Sequence, Union

from xir.decimal_complex import DecimalComplex
from xir.program import Declaration, ObservableStmt, Param, Program, Statement, Wire

VALID_CONSTANTS = {"PI"}


class ValidationError(Exception):
    """Exception raised when validation of an XIR program fails.

    Args:
        validation_msgs (Sequence, str): A Sequence of error messages describing which validators
            failed or a single custom message. If ``None`` then a generic error message is used.
    """

    def __init__(self, validation_msgs: Union[None, Sequence, str] = None) -> None:
        if isinstance(validation_msgs, str):
            self.message = f"XIR program is invalid: {validation_msgs}"

        elif isinstance(validation_msgs, Sequence) and validation_msgs:
            validations_str = "\n\t-> " + "\n\t-> ".join(map(str, validation_msgs))
            self.message = (
                "XIR program is invalid: the following issues have been detected:"
                f"{validations_str}"
            )

        else:  # backwards compatible with string messages
            self.message = "XIR program is invalid."

        super().__init__(self.message)


def stem(val: str) -> str:
    """Return stem of function or observable call by removing any terminal parantheses.

    As an example, passing ``"function(param)"`` returns ``"function"``.

    Args:
        val (str): function call string

    Returns:
        str: stem of input string
    """
    match = re.match(r"(\w+)\s*\(.*?\)\s*$", val)
    if match:
        return match[1]
    return val


class Validator:
    """Validator used to validate an XIR program.

    Checks if an XIR program is correctly defined. While the XIR parser and transformer catch
    syntactic errors, the validator ensures that all statements are logically valid and consistent.

    Args:
        program (Program): the program which is to be validated
        ignore_includes (bool): whether to ignore any include statements and always check that all
            statements have their corresponding declarations in the same file

    **Example:**

    The ``Validator`` class is used to create a validator object which accepts a program to
    be validated. The validator can then be run by calling the ``run()`` method. Optionally,
    the user can set the ``raise_exception`` flag, which determines whether or not to raise a
    ``ValidationError`` or return ``None``, or to simply return the error message log without
    raising an exception.

    .. code-block:: python

        valid_script = \"""
            gate rx(a)[0];
            out amplitude(state)[0..2];

            rx(0.7) | [0];
            rx(1.2) | [1];

            amplitude(state: [1, 0]) | [0..2];
        \"""

        prog = parse_script(valid_script)
        xir.Validator(prog).run()

    Optionally, the validator can be set to ignore any include statements, meaning that all gates,
    outputs, functions and observables must be declared in the same file.

    .. code-block:: python

        validator = xir.Validator(prog, ignore_includes=True)
        log = validator(raise_exception=False).run()
    """

    def __init__(self, program: Program, ignore_includes: bool = False) -> None:
        self._validation_messages = []
        self._ignore_includes = ignore_includes

        self._program = program
        self.has_includes = len(program.includes) > 0

        self._validators = {
            "constants": True,
            "declarations": True,
            "statements": True,
            "definitions": True,
        }

    def run(self, raise_exception: bool = True) -> Optional[Sequence[str]]:
        """Runs the validation checks.

        Args:
            raise_exception (bool): whether to raise a ``ValidationError``
                if any issues are found

        Returns:
            Sequence[str], None: list of potential issues iff ``raise_exception`` is ``False``
            and at least one validation error was detected

        Raises:
            ValidationError: if any issues are found and ``raise_exception`` is ``True``
        """
        # reset validation messages in case previously run
        self._validation_messages = []

        if self._validators["constants"]:
            self._check_constants()

        if self._validators["declarations"]:
            self._check_declarations()

        if self._validators["statements"]:
            self._check_statements()

        if self._validators["definitions"]:
            self._check_recursive_definitions()
            self._check_gate_definitions()
            self._check_observable_definitions()

        if self._validation_messages and raise_exception:
            raise ValidationError(self._validation_messages)

        return self._validation_messages or None

    def _check_constants(self) -> None:
        """Checks that the declared constants are valid."""
        for const in self._program.constants.keys():
            # grammar uses lowercase constants, while the program uses upper-case
            if const in map(str.lower, VALID_CONSTANTS):
                msg = f"Constant '{const}' is already defined and cannot be replaced."
                self._validation_messages.append(msg)

    def _check_declarations(self) -> None:
        """Checks that declarations are valid.

        Checks that all declarations have correct wires, labels and parameters.
        The declarations will be marked as invalid if:

            * A declaration has duplicate wire labels.
            * A declaration has duplicate parameter names.
            * A declaration has labels which are not strings.
            * A declaration has parameters which are not strings.
        """
        for decl in (d for l in self._program.declarations.values() for d in l):
            if len(set(decl.wires)) != len(decl.wires):
                msg = f"Declaration '{decl}' has duplicate wires labels."
                self._validation_messages.append(msg)

            if len(set(decl.params)) != len(decl.params):
                msg = f"Declaration '{decl}' has duplicate parameter names."
                self._validation_messages.append(msg)

            if not all(isinstance(p, str) for p in decl.params):
                msg = f"Declaration '{decl}' has parameters which are not strings."
                self._validation_messages.append(msg)

    def _check_statements(
        self,
        statements: Optional[Sequence[Statement]] = None,
        declared_params: Optional[Sequence[str]] = None,
    ) -> None:
        """Checks that statements are valid.

        Checks that all statements have a declaration and that they are correctly applied. If no
        statements are passed, the script-level statements in the program will be used. The
        statements will be marked as invalid if:

            * A gate application statement specifies the wrong number of wires.
            * A gate application statement specifies the wrong number of parameters.
            * A gate application statement specifies the wrong parameter names.
            * A gate application statement at the script level is applied to named wires.

        Checks that modifiers such as ``ctrl`` and ``inv`` are correctly applied.
        The modifiers will be marked as invalid if:

            * An ``inv`` or ``ctrl`` modifier is applied to an output statement.
            * There are duplicate wires specified in a ``ctrl`` modifier.

        If ``declared_params`` are passed, parameters may have the containing names.
        """
        declarations = list(
            chain(self._program.declarations["gate"], self._program.declarations["out"])
        )
        declaration_names = [x.name for x in declarations]

        statements = statements or self._program.statements
        for stmt in statements:
            if stmt.name in declaration_names:
                self._check_statements_with_declarations(stmt, declarations, declaration_names)

            elif self._ignore_includes or not self.has_includes:
                msg = f"Name '{stmt.name}' has not been declared."
                self._validation_messages.append(msg)

            self._check_statement_applications(stmt, declared_params)

    def _check_statements_with_declarations(
        self, stmt: Statement, declarations: Sequence[Declaration], declaration_names: Sequence[str]
    ) -> None:
        """Checks that statements comply with their declarations."""
        idx = declaration_names.index(stmt.name)

        # check that dict-like parameters use the correct parameter names
        if isinstance(stmt.params, dict):
            if set(stmt.params.keys()) != set(declarations[idx].params):
                expected = ", ".join(declarations[idx].params)
                msg = f"Statement '{stmt}' passes the wrong parameters. Expected '{expected}'."
                self._validation_messages.append(msg)

        # check that the number of parameters agree with the declaration
        # redundant if prior error is present
        elif len(stmt.params) != len(declarations[idx].params):
            expected = len(declarations[idx].params)
            msg = f"Statement '{stmt}' has {len(stmt.params)} parameter(s). Expected {expected}."
            self._validation_messages.append(msg)

        # check that statements are applied to the correct number of wires
        if len(stmt.wires) != len(declarations[idx].wires):
            expected = len(declarations[idx].wires)
            msg = f"Statement '{stmt}' has {len(stmt.wires)} wire(s). Expected {expected}."
            self._validation_messages.append(msg)

        # check that gates aren't applied to duplicate wires
        if len(set(stmt.wires)) != len(stmt.wires):
            msg = f"Statement '{stmt}' is applied to duplicate wires."
            self._validation_messages.append(msg)

        # check that `inv` or `ctrl` isn't applied to an output statement
        if (stmt.is_inverse or stmt.ctrl_wires) and declarations[idx].type_ == "out":
            msg = f"Statement '{stmt}' is an output statement but has 'ctrl' or 'inv' modifiers."
            self._validation_messages.append(msg)

    def _check_statement_applications(
        self, stmt: Statement, declared_params: Optional[Sequence[str]]
    ) -> None:
        """Checks that statements are valid.

        Checks the remaining conditions documented (but not implemented) by
        :func:`Validator._check_statements()`.
        """
        # check that only integer wire labels are used at the script level
        if declared_params is None and any(isinstance(wire, str) for wire in stmt.wires):
            msg = (
                f"Statement '{stmt}' is applied to named wires. Only integer wire labels are "
                "allowed at the script level."
            )
            self._validation_messages.append(msg)

        # check that control wires aren't the same as applied wires
        ctrl_stmt_wires = set(stmt.ctrl_wires) & set(stmt.wires)
        if ctrl_stmt_wires:
            msg = f"Statement '{stmt}' has control wires {ctrl_stmt_wires} which are also applied."
            self._validation_messages.append(msg)

    def _check_recursive_definitions(self) -> None:
        """Checks that a gate is not defined in terms of itself."""

        def _recursive_check(name: str, definition: Sequence[Statement], stack: MutableSet[str]):
            if name in stack:
                msg = f"Gate definition '{name}' has a circular dependency."
                self._validation_messages.append(msg)
                return

            stack.add(name)

            for stmt in definition:
                if stmt.name in self._program.gates.keys():
                    _recursive_check(stmt.name, self._program.gates[stmt.name], stack)

            stack.remove(name)

        for name, val in self._program.gates.items():
            _recursive_check(name, val, stack=set())

    def _check_gate_definitions(self) -> None:
        """Checks that gate definitions are valid.

        Definitions are invalid if:
            * A non-integer wire label is used which is not explicitly specified
              in the gate signature.
            * At least one wire label is specified in the gate signature and an
              integer wire label is used which is not explicitly specified in
              the gate signature.
            * A mixture of integer and non-integer wire labels are specified in
              the gate signature.
            * A wire label which is specified in the gate signature is not used.
            * The same name is used for wires and parameters.
            * The statements inside gate definitions are not correctly declared.
        """
        for name, statements in self._program.gates.items():
            # check that statements in definitions are valid
            declared_params = self._program.search("gate", "params", name)
            self._check_statements(statements, declared_params)

            applied_wires = list(
                chain(*map(lambda x: list(x.wires) + list(x.ctrl_wires), statements))
            )

            declared_wires = self._program.search("gate", "wires", name)

            self._shared_definition_checks(name, declared_wires, applied_wires, declared_params)

    def _check_observable_definitions(self) -> None:
        """Check that observable definitions are valid.

        Definitions are invalid if:
            * A non-integer wire label is used which is not explicitly specified
              in the observable signature.
            * At least one wire label is specified in the observable signature
              and an integer wire label is used which is not explicitly
              specified in the observable signature.
            * A mixture of integer and non-integer wire labels are specified in
              the observable signature.
            * A wire label which is specified in the gate signature is not used.
            * The same name is used for wires and parameters.
            * The statements inside observable definitions are not correctly declared.
        """
        for name, statements in self._program.observables.items():
            declared_params = self._program.search("obs", "params", name)

            declared_wires = self._program.search("obs", "wires", name)
            applied_wires = list(chain(*[x.wires for x in statements]))

            self._check_observable_statements(statements, declared_params)
            self._shared_definition_checks(name, declared_wires, applied_wires, declared_params)

    def _shared_definition_checks(
        self,
        name: str,
        declared_wires: Sequence[Wire],
        applied_wires: Sequence[Wire],
        declared_params: Sequence[Param],
    ) -> None:
        """Checks the names, wires, and parameters for both observable and gate definitions."""
        any_integer_applied_wires = any(isinstance(w, int) for w in applied_wires)
        all_string_declared_wires = all(isinstance(w, str) for w in declared_wires)

        if any_integer_applied_wires and all_string_declared_wires:
            msg = (
                f"Definition '{name}' is invalid. Only named wires can be applied when "
                "declaring named wires."
            )
            self._validation_messages.append(msg)

        if set(applied_wires) - set(declared_wires):
            applied = ", ".join(map(str, applied_wires))
            declared = ", ".join(map(str, declared_wires))
            msg = (
                f"Definition '{name}' is invalid. Applied wires [{applied}] differ "
                f"from declared wires [{declared}]."
            )
            self._validation_messages.append(msg)

        if set(declared_wires) & set(declared_params):
            msg = f"Definition '{name}' is invalid. Wire and parameter names must differ."
            self._validation_messages.append(msg)

        constants_declared = set(declared_params) & set(self._program.constants)
        if constants_declared:
            msg = (
                f"Definition '{name}' is invalid. Cannot use declared constant(s) "
                f"{constants_declared} as parameter(s)."
            )
            self._validation_messages.append(msg)

    def _check_observable_statements(
        self,
        statements: Sequence[ObservableStmt],
        declared_params: Optional[Sequence[str]] = None,
    ) -> None:
        """Check that observable statements are valid.

        The statements will be marked as invalid if:

            * There is an undeclared prefactor variable.
            * An invalid observable is used.
            * A product of observables is applied to the same wire(s).
        """
        declared_funcs = [decl.name for decl in self._program.declarations["func"]]

        for stmt in statements:
            declared_pref = (
                stmt.pref in declared_params
                or str(stmt.pref) in VALID_CONSTANTS
                or (isinstance(stmt.pref, str) and stem(stmt.pref) in declared_funcs)
                or isinstance(stmt.pref, (int, float, complex, Decimal, DecimalComplex))
            )
            if not declared_pref and (self._ignore_includes or not self.has_includes):
                msg = f"Statement '{stmt}' has an undeclared prefactor variable '{stmt.pref}'."
                self._validation_messages.append(msg)

            # check if invalid observables are applied
            words, wires = zip(*stmt.terms)
            invalid_words = set(words) - {x.name for x in self._program.declarations["obs"]}
            if invalid_words and not (not self._ignore_includes and self.has_includes):
                msg = (
                    f"Observable statement '{stmt}' is invalid. Observable(s) "
                    f"{sorted(invalid_words)} have not been declared."
                )
                self._validation_messages.append(msg)

            # check if observable products are applied to the same wires
            if len(set(wires)) != len(wires):
                msg = (
                    f"Observable statement '{stmt}' is invalid. Products of observables cannot be "
                    "applied to the same wires."
                )
                self._validation_messages.append(msg)



from ._version import __version__
from .decimal_complex import DecimalComplex
from .parser import Transformer, parser
from .program import Declaration, ObservableStmt, Program, Statement
from .validator import Validator

__all__ = ["DecimalComplex", "Declaration", "ObservableStmt", "Program", "Statement", "Transformer"]


def parse_script(script: str, **kwargs) -> Program:
    """Parses an XIR script into a structured :class:`xir.Program`."""
    tree = parser.parse(script)
    return Transformer(**kwargs).transform(tree)
